<copyright> TESLA mu.Extension extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Extension.t,v 1.28 1999/09/30 07:19:09 tiggr Exp $</id>
    </copyright>

implementation class
Extension: tesla, posing mu.Extension
{
  <doc> Boolean flags to store in the {int} {flags}, so we do not spend a
      whole {boolean} (the size of a byte) on each one.  </doc>
  const EXT_IMPLEMENTED = 1;
}

end;

implementation instance
Extension
{
  <doc> Metas directly introduced by this extension.  </doc>
  MutableArray metas;

  <doc> The load method of this extension.  </doc>
  mutable Method load_method;

  <doc> The alignment required by the variables for the state of this
      extension.  </doc>
  int variables_alignment;

  <doc> Flags.  </doc>
  int flags;
}

<doc> Note that this extension is responsible for inheriting the {meta}.
    </doc>
void
  addMeta Meta meta
{
  if (!metas)
    metas = [MutableObjectArray with meta];
  else
    [metas add meta];
}

<doc> Issue an error if we are adding non-static state to a class in
    an extension in another unit than the one containing the class.
    </doc>
void
  addVariable MetaVariable v
post
  variables_alignment <= MAX_ALIGNMENT
{
  if (![self isMainExtension] && [meta isClass] && ![v isStatic]
      && [self unit] != [meta unit])
    [[[ParseTom error v]
       print "unimplemented: non-static class variables in another unit"] nl];

  int align = [[v type] requiredAlignment];
  if (align > variables_alignment)
    variables_alignment = align;

  [super addVariable v];
}

<doc> Do everything necessary after all units have been loaded.  </doc>
void
  allUnitsLoaded
{
}

<doc> Return {TRUE} if the {file} of this extension is the
    {builtin_file}.  </doc>
boolean
  builtin
{
  = file == [TomFile builtin_file];
}

<doc> Add the {load_method} if the {unit} is our unit.  </doc>
MutableArray (collect)
  collectLoadMethods MutableArray collect
	     forUnit mu.Unit unit
{
  if (load_method != nil && [file unit] == unit)
    [collect add load_method];
}

<doc> Return whether {EXT_IMPLEMENTED} is set in the {flags}.  </doc>
boolean
  implemented
{
  = !!(flags & EXT_IMPLEMENTED);
}

void
  makeMethodsPerform tom.Invocation inv
{
  int i, n = [methods length];
  for (i = 0; i < n; i++)
    [inv fireAt methods[i]];
}

<doc> Add to {collect}, created when incoming is {nil}, those methods that
    fit the {name_parts} with the {actuals}, as requested by the {sender}.
    </doc>
MutableArray
   methods MutableArray collect
     named Indexed name_parts
      with Indexed actuals
       for Expression sender
{
  if (methods != nil)
    {
      int i, n = [methods length];
      for (i = 0; i < n; i++)
	{
	  CustomMethod m = methods[i];
	  if ([m namePartsMatch name_parts withActuals actuals])
	    {
	      // confine sender visibility
	      if (!collect)
		collect = [MutableObjectArray with m];
	      else
		[collect add m];
	    }
	}
    }
  = collect;
}

<doc> Set or clear {EXT_IMPLEMENTED} in the {flags}.  </doc>
void
  set_implemented boolean imp_p
{
  flags = imp_p ? flags | EXT_IMPLEMENTED : flags & ~EXT_IMPLEMENTED;
}

<doc> Do everything that can (and should) be done at the end of loading a
    single unit.  </doc>
void
  singleUnitLoaded boolean second_pass
pre
  !current_meta
post
  !current_meta
{
  /* First let the corresponding extension of the class do this.  */
  if ([meta isInstance])
    [[[meta itsClass] extensionNamed name] singleUnitLoaded second_pass];

  // tomCCC We'd like to reset the current_meta with an unwind, but return
  // from an unwind is not properly implemented...
  // Tue Apr 28 15:23:49 1998, tiggr@gerbil.org
  current_meta = meta;
  if (second_pass)
    {
      if (![self builtin])
	{
	  /* Second pass.  */
	  if (other_attributes != nil)
	    {
	      int i, n = [other_attributes length];
	      for (i = 0; i < n; i++)
		[other_attributes[i] singleUnitLoaded second_pass];
	    }

	  if (methods != nil)
	    {
	      int i, n = [methods length];
	      for (i = 0; i < n; i++)
		[CustomMethod (methods[i]) singleUnitLoaded 1];
	    }
	}

      current_meta = nil;
      return;
    }

  /* First pass.  */
  if (![self implemented] && ![self builtin])
    {
      [[[ParseTom error self]
	 print ("file ", [file filename], " fails to implement extension ",
		[self descriptive_name])] nl];
      current_meta = nil;
      return;
    }

  /* Tell the meta that the unit was loaded if we're the main extension.  */
  if ([self isMainExtension])
    [meta singleUnitLoaded second_pass];

  /* Inform the attributes.  Aliases will find the original, public or
     mutable ones will define the corresponding methods, etc.  */
  if (other_attributes != nil)
    {
      int i, n = [other_attributes length];
      for (i = 0; i < n; i++)
	[other_attributes[i] singleUnitLoaded second_pass];
    }
  if (variables != nil)
    {
      int i, n = [variables length];
      for (i = 0; i < n; i++)
	[variables[i] singleUnitLoaded second_pass];
    }

  if (methods != nil && ![self builtin])
    {
      int i, n = [methods length];
      for (i = 0; i < n; i++)
	[CustomMethod (methods[i]) singleUnitLoaded 0];
    }

  current_meta = nil;

  [meta singleUnitLoaded self];
}

<doc> Write something like {_ei_i_tom_All}.  </doc>
OutputStream (s)
  writeCExtensionIdentity OutputStream s
{
  [self writeCNameSuffix [s print "_ei_"]];
}

<doc> Write something like {&_ed_i_tom_All,}.  </doc>
OutputStream (s)
  writeCExtensionList OutputStream s
	      forUnit mu.Unit unit
{
  if ([file unit] == unit)
    [[self writeCNameSuffix [s print "&_ed_"]] print ",\n"];
}

<doc> Write something like {_es_i_tom_All}.  </doc>
OutputStream (s)
  writeCExtensionStructName OutputStream s
{
  [self writeCNameSuffix [s print "_es_"]];
}

<doc> Write something like {_ep_i_tom_All}.  </doc>
OutputStream (s)
  writeCNameStatePointer OutputStream s
{
  [self writeCNameSuffix [s print "_ep_"]];
}

TeslaOutput (s)
  writeCNameStatePointerDeclaration TeslaOutput s
				for Meta m
{
  String whoami = ([m isInstance] && ![meta isInstance]
		   ? "(((tom_object) self)->isa)" : "((tom_object) self)");

  [[self writeCExtensionStructName [[s nl] print "struct "]] print " *"];
  [[self writeCNameStatePointer s] print (" = (void *) ((char *) ", whoami,
					  " + ", whoami, "->isa->info.eot->")];
  [[self writeCExtensionIdentity [s print "offset["]] print "]);"];
}

<doc> Write something like {_ep_i_tom_All}.  </doc>
OutputStream (s)
  writeCNameStatePrefix OutputStream s
{
  [[self writeCNameStatePointer s] print "->"];
}

<doc> Write something like {i_tom_All}.  </doc>
OutputStream (s)
  writeCNameSuffix OutputStream s
{
  [meta writeCReferenceTypeName s];
  if (name != nil)
    [s print ("_", name)];
}

<doc> Compile.  </doc>
TeslaOutput (of)
  writeCFile TeslaOutput of
{
  /* First let the corresponding extension of the class do this.  */
  if ([meta isInstance])
    [[[meta itsClass] extensionNamed name] writeCFile of];

  int num_methods, num_statics;
  unwind (current_meta = nil)
    ({
      current_meta = meta;

      [[self writeCNameSuffix [of print "\n/* Start of extension "]]
	print ".  */\n"];

      if (variables != nil)
	[[self writeCExtensionIdentity [of print "unsigned int "]]
	  print " = 0;\n"];

      if (![self builtin])
	{
	  if (methods != nil)
	    {
	      int i, n = [methods length];
	      for (i = 0; i < n; i++)
		{
		  if ([methods[i] hasImplementation])
		    num_methods++;
		  [CustomMethod (methods[i]) writeCFile of];
		}
	      [of print '\n'];
	    }

	  /* Define the static class variables and their description.  */
	  if (other_attributes != nil)
	    {
	      int i, n = [other_attributes length];
	      MutableArray statics;

	      for (i = 0; i < n; i++)
		statics = [other_attributes[i] writeDefineStaticVariables of
					   collect statics];

	      if (statics != nil)
		{
		  num_statics = [statics length];
		  [[self writeCNameSuffix
			 [of print "struct trtd_static_var _static_ed_"]]
		    print "[] = {"];
		  for (i = 0; i < num_statics; i++)
		    [statics[i] writeCStaticDescription of];
		  [of print "};\n"];
		}
	    }
	}
    });

  /* Describe the methods.  */
  if (num_methods > 0)
    {
      [[self writeCNameSuffix
	     [of print "\nstatic struct trtd_methods trtd_methods_"]]
	print (" = {", num_methods, ", {")];

      int i, n = [methods length];
      for (i = 0; i < n; i++)
	[methods[i] writeCMethodDescription of];
      [of print "}};\n"];
    }

  /* Describe the metas introduced by this extension.  Always make it
     inherit the State class.  */
  if (state_class != nil && [self isMainExtension] && [meta isClass])
    if (!metas)
      metas = [MutableObjectArray with state_class];
    else
      [metas add state_class];
  if (metas != nil)
    {
      int i, n = [metas length];
      [[self writeCNameSuffix [of print "static struct trt_metas metas_"]]
	print (" = {{", n, ", 0}, {")];
      for (i = 0; i < n; i++)
	{
	  Meta m = metas[i];
	  [[m writeCMetaDefinitionName [of print "\n&"]] print ','];
	}
      [of print "\n}};\n"];
    }

  /* Describe the extension.  */
  [[self writeCNameSuffix [of print "struct trtd_extension _ed_"]]
    print " = {"];
  {
    /* struct trt_class *meta; */
    [[meta writeCMetaDefinitionName [of print "\n&"]] print ','];

    /* tom_object extension_object; */
    [of print "\n0,"];

    /* int *eid_in_a_global; */
    if (!variables)
      [of print "\n0,"];
    else
      [[self writeCExtensionIdentity [of print "\n&"]] print ','];

    /* struct name name; */
    if (!name)
      [of print "\n{0, 0},"];
    else
      [of print ("\n{\"", name, "\", ", [name length], "},")];

    /* struct trtd_methods *methods; */
    if (num_methods > 0)
      [[self writeCNameSuffix [of print "\n&trtd_methods_"]] print ','];
    else
      [of print "\n0,"];

    /* struct trt_metas *supers; */
    if (!metas)
      [of print "\n0,"];
    else
      [[self writeCNameSuffix [of print "\n&metas_"]] print ','];

    /* int state_size; */
    if (!variables)
      [of print "\n0,"];
    else
      [[self writeCExtensionStructName [of print "\nsizeof (struct "]]
	print "),"];

    /* int state_align; */
    [of print ("\n", variables_alignment, ",")];

    /* struct trtd_static_var *statics;
       int num_statics; */
    if (!num_statics)
      [of print "\n0, 0,"];
    else
      [[self writeCNameSuffix [of print "\n_static_ed_"]]
	print (", ", num_statics, ",")];

    /* normal variables description.  */
    if (!variables)
      [of print "\n0"];
    else
      {
	int i, n = [variables length];

	/* int num_vars; */
	[of print ('\n', n, ",{")];

	/* struct trtd_normal_var vars[0]; */
	for (i = 0; i < n; i++)
	    [variables[i] writeCNormalDescription of];

	[of print '}'];
      }
  }
  [of print "};\n"];

  [[self writeCNameSuffix [of print "/* End of extension "]] print ".  */\n"];
}

void
  writeRCFile OutputStream of
	 pass int pass
      collect MutableArray collect
{
  /* First let the corresponding extension of the class do this.  */
  if ([meta isInstance])
    [[[meta itsClass] extensionNamed name]
      writeRCFile of pass pass collect collect];
  if (pass == 0)
    {
      [collect add self];

      [[self writeCNameSuffix [of print "extern struct trtd_extension _ed_"]]
	print ";\n"];
    }
}

<doc> Write the info on this extension to the RH file.  </doc>
void
   writeRHFile OutputStream of
	  pass int pass
{
  /* First let the corresponding extension of the class do this.  */
  if ([meta isInstance])
    [[[meta itsClass] extensionNamed name] writeRHFile of pass pass];

  if (pass == 0)
    {
      if (variables != nil)
	{
	  /* Declare the struct carrying this extension's state, and its
	     identity.  */
	  [self writeCExtensionIdentity [of print "extern unsigned int "]];
	  [[self writeCExtensionStructName [of print ";\nstruct "]] print " {"];
	  int i, n = [variables length];
	  for (i = 0; i < n; i++)
	    {
	      MetaVariable v = variables[i];
	      [v writeCDeclaration [of print ' ']];
	    }

	  if (meta == state_class && [self isMainExtension])
	    [of print " struct trt_class_info info;"];
	  [of print " };\n"];
	}

      if (other_attributes != nil)
	{
	  /* Declare the static variables.  */
	  int i, n = [other_attributes length];
	  for (i = 0; i < n; i++)
	    [other_attributes[i] writeCDeclaration of];
	}
    }
}

end; 
