<copyright> TESLA mu extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: various.t,v 1.38 1999/09/25 22:25:43 tiggr Exp $</id>
    </copyright>

/******************** Alias ********************/

implementation class
Alias extension tesla: tesla

end;

implementation instance
Alias extension tesla

<doc> Return a cast of the original.  </doc>
Cexpr
    precompile Cblock parent
	method Cbody body
{
  Cexpr r = [original precompile parent method body];
  = [[Ccast alloc] init (type, r)];
}

<doc> Find the original object being redeclared by this alias.  </doc>
void
  singleUnitLoaded boolean second_pass
// pre
//   && !original
{
  if (!second_pass)
    {
      original
	= MetaVariable ([[ext meta] attributeNamed name
				    including (FALSE, TRUE, FALSE, TRUE,
					       ![ext isMainExtension], FALSE)
				    excluding: self]);
      if (!original)
	[[[ParseTom error self] print "REDECLARE of undeclared original"] nl];
    }
}

<doc> Do nothing.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
}

end;

/******************** ArrayReference ********************/

implementation class
ArrayReference extension tesla: tesla

end;

implementation instance
ArrayReference extension tesla

<doc> Return a replacement for the {assignment} by turning it into an
    invocation of {set at}.  </doc>
Expression
  replacementForAssignmentBy Assignment assignment
{
  = [[UnresolvedInvocation alloc assignment]
      initWithReceiver (array, nil, FALSE)
      nameParts [MutableObjectArray with ([UniqueString with "set"],
					  [UniqueString with "at"])]
      arguments [MutableObjectArray with ([assignment rhs], index)]];
}

<doc> Assume the type of the {referenced}, and pass to {super}.  </doc>
(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, indices_index)
{
  Expression rhs
    = [[UnresolvedInvocation alloc self]
	initWithReceiver (array, nil, FALSE)
	nameParts [MutableObjectArray with [UniqueString with "at"]]
	arguments [MutableObjectArray with index]];

  return [rhs resolve (expected, convertible) context context_type
	      nested: (indices, indices_index)];
}

end;

/******************** Attribute ********************/

implementation class
Attribute extension tesla: tesla

end;

implementation instance
Attribute extension tesla

<doc> Be informed that the unit of the containing extension has been
    loaded.  Default implementation does nothing.  </doc>
void
  singleUnitLoaded boolean second_pass
{
}

deferred Expression
  reference Instance context_instance;

<doc> If the receiving attribute is a static meta variable, add it to
    the array {a}, creating it when needed, and emit the definition of
    the variable to the stream {s}.  Return {a}, or the newly created
    array.  The default implementation simply returns what has been
    collected so far, i.e., {a}.  </doc>
MutableArray (a)
  writeDefineStaticVariables TeslaOutput s
		     collect MutableArray a
{
}

end;

/******************** Break ********************/

implementation class
Break: Expression

end;

implementation instance
Break
{
  <doc> The loop which we break (or something).  </doc>
  Loop loop;

  <doc> The value.  </doc>
  Expression value;
}

<doc> Designated initializer.  </doc>
id
  init (Expression, Loop) (v, l)
{
  (value, loop) = (v, l);
  = [super initWithType nil];
}

<doc> Return the name of this statement in C.  Default returns the string
    {break}.  </doc>
String
  CGotoName
{
  = "break";
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  [value coarseUsage f | EXPR_RHS parent c body m];
}

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  if (value != void_expr)
    {
      Cexpr r = [loop cresult];
      Cexpr v = [value precompile parent method body];
      [r precompile parent method body assignmentOf v];
    }

  [parent add [[Cstmt alloc] init [self CGotoName]]];
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    {
      if (value != void_expr)
	value = [value resolve ([Cons with ([loop type], nil)], nil)
		       context context_type];
      type = [value force_type];
    }
  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

end;

/******************** Cast ********************/

implementation class
Cast extension tesla: tesla

end;

implementation instance
Cast extension tesla

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  [expr coarseUsage f parent c body m];
}

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  Cexpr e = [expr precompile parent method body];
  result = [type temporaryCexpr parent];
  [parent add [[Casgn alloc] init (result, e)]];
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    {
      expr = [expr resolve (nil, desired_type) context context_type];
      // Check for a valid cast?
      // Mon May 11 11:18:57 1998, tiggr@gerbil.org
      [expr force_type];
      type = desired_type;
    }
  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

end;

/******************** CharConstant ********************/

implementation class
CharConstant extension tesla: tesla

end;

implementation instance
CharConstant extension tesla

<doc> Return {TRUE}.  </doc>
boolean
  isConstant
{
  = TRUE;
}

Cexpr
  precompile Cblock parent
      method Cblock body
{
  MutableByteString buf = [MutableByteString withCapacity 4];
  [self writeCFile buf];

  = [[Cconst alloc self] init buf];
}

OutputStream (s)
  writeCFile OutputStream s
{
  if (value == '\'' || value == '\\' || value < ' ' || value > '~')
    {
      [s print "'\\"];
      if (value == '\'' || value == '\\')
	[s print value];
      else if (value == 10)
	[s print "n"];
      else
	[s print (int (value / 64), int (value % 64) / 8, int (value % 8))];
      [s print "'"];
    }
  else
    [s print ('\'', value, '\'')];
}

end;

/******************** ClassReference ********************/

implementation class
ClassReference extension tesla: tesla

end;

implementation instance
ClassReference extension tesla

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  // Naughty!
  // Mon May 25 19:01:04 1998, tiggr@gerbil.org
  = Any (self);
}

TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [Meta (type) writeCReferenceTypeName [s print "_mr_"]];
}

end;

/******************** Constant ********************/

implementation class
Constant extension tesla

end;

implementation instance
Constant extension tesla

<doc> Do nothing, really.  </doc>
void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  if (var_flags & CONSTANT_DID_USAGE == 0)
    {
      var_flags |= CONSTANT_DID_USAGE;
      // Not really correct...  pray that f includes EXPR_RHS if needed
      // Fri May 29 12:21:17 1998, tiggr@gerbil.org
      [value coarseUsage f parent c body m];
    }
}

<doc> If we do not have a type yet, resolve the {value} and get its
    {type}.  </doc>
Type
  force_type
{
  if (!type)
    {
      value = [value resolve (nil, nil) context [ext meta]];
      type = [value force_type];
    }
  = type;
}

Cexpr
    precompile Cblock parent
	method Cblock body
{
  = [value precompile parent method body];
}

Expression
  reference Instance context_instance
{
  = [[Reference alloc] initWithReferenced self fromInstance context_instance];
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    [self force_type];
  = [value resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

<doc> Resolve the {value}.  </doc>
void
  singleUnitLoaded boolean second_pass
{
  if (second_pass)
    value = [value resolve (nil, nil) context current_meta];
}

<doc> If the {value} of this constant is constant, emit a {#define}
    for it.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
  if ([value isConstant])
    {
      [s print "#define "];
      [C write [[ext writeCNameSuffix s] print '_'] CName name];
      [[value writeCFile [s print ' ']] print '\n'];
    }
}

end;

/******************** Continue ********************/

implementation class
Continue: Break

end;

implementation instance
Continue

<doc> Return a {goto label}.  </doc>
String
  CGotoName
{
  = [[MutableByteString new] print ("goto ", [loop continue_label])];
}

end;

/******************** ITE ********************/

implementation class
ITE: Expression

end;

implementation instance
ITE
{
  <doc> The deciding condition.  </doc>
  Expression condition;

  <doc> The alternatives.  </doc>
  Expression then_part, else_part;
}

<doc> Designated initializer.  </doc>
id
  initWithCondition Expression c
	   branches (Expression, Expression) (tp, ep)
{
  (then_part, else_part) = (tp, ep);
  condition = c;

  = [super initWithType nil];
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  [condition coarseUsage EXPR_RHS parent c body m];
  [then_part coarseUsage f parent c body m];
  if (else_part != nil)
    [else_part coarseUsage f parent c body m];
}

<doc> Inform our elements.  </doc>
TeslaOutput (of)
  compileBlocks TeslaOutput of
{
  [condition compileBlocks of];
  [then_part compileBlocks of];
  if (else_part != nil)
    [else_part compileBlocks of];
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (type != nil)
    = [super resolve (expected, convertible) context context_type
	     nested: (indices, index)];

  Cons pos_t, pos_e, pos;
  Expression e;

  if (![condition type])
    condition = [condition resolve
		  ([Cons with (basic_types_from_kind[BT_BOOLEAN], nil)],
		   nil) context context_type];
  (then_part, pos_t) = [then_part resolve (expected, convertible)
				  context context_type
				  nested: (indices, index)];
  if (!pos_t)
    return (self, nil);
  if (!else_part)
    pos = pos_t;
  else
    {
      (else_part, pos_e) = [else_part resolve (expected, convertible)
				      context context_type
				      nested: (indices, index)];
      if (!pos_e)
	return (self, nil);
      pos = [Type intersectTypes (pos_t, pos_e) expected expected];
    }

  if (pos != nil)
    {
      if (![pos cdr])
	{
	  /* We are fully defined (or should be).  */
	  Type t = [pos car];

	  if (t == dynamic_type)
	    pos = nil;

	  if (![then_part type])
	    then_part = [then_part resolve (pos, convertible)
				   context context_type
				   nested: (indices, index)];
	  if (else_part != nil && ![else_part type])
	    else_part = [else_part resolve (pos, convertible)
				   context context_type
				   nested: (indices, index)];

	  if (t == dynamic_type)
	    {
	      // Wrong?
	      type = [then_part force_type];
	    }
	  else
	    type = [t typeAt index in indices];
	}
    }
  else if (convertible != nil || expected != nil)
    {
      // What if EXPECTED contains void?
      // Mon May 11 12:15:28 1998, tiggr@gerbil.org
      [[[ParseTom error self] print "unable to unify branch types"] nl];
    }
  else
    {
      /* The types of the branches do not match, but we can live
	 with that: our type becomes void.  */
      type = void_type;
      if (!expected)
	pos = [Cons with (type, nil)];
      else
	{
	  Type t = [type elementOf expected covariant YES
			 context context_type nested: (indices, index)];
	  pos = [Cons with (t, nil)];
	}

      if (![then_part type])
	then_part = [then_part resolve (nil, nil) context context_type];
      if (else_part != nil && ![else_part type])
	else_part = [else_part resolve (nil, nil) context context_type];
    }

  = (type != nil ? [super resolve (expected, convertible) context context_type
			  nested: (indices, index)]
     : (self, pos));
}

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  Cblock eb, tb = [[Cblock alloc] initWithParent parent];

  if (expr_flags & EXPR_RHS != 0)
    result = [type temporaryCexpr parent];
  if (result != nil || else_part != nil)
    eb = [[Cblock alloc] initWithParent parent];

  Cexpr condition_expr = [condition precompile parent method body];
  Cexpr then_expr = [then_part precompile tb method body];
  if (result != nil)
    [then_expr compile tb assignmentTo result];

  Cexpr else_expr;
  if (else_part != nil)
    else_expr = [else_part precompile eb method body];
  else if (result != nil)
    else_expr = [[type defaultValue result] precompile eb method body];

  if (result != nil)
    [else_expr compile eb assignmentTo result];

  [parent add [[Cite alloc condition] init (condition_expr, tb, eb)]];
}

end;

/******************** ForeignCode ********************/

<doc> {ForeignCode} is a {ForeignExpression} which has a side effect and
    therefor is emitted (presumably in a {void} context.  </doc>
implementation class
ForeignCode: ForeignExpression

end;

implementation instance
ForeignCode

<doc> Designated initializer.  </doc>
id
  initWithExpression String e
{
  = [super initWithExpression e type basic_types_from_kind[BT_VOID]];
}

<doc> Precompilation when at the top level.  </doc>
Cexpr
  precompile
{
  = [[Cconst alloc self] init expression];
}

<doc> Precompilation when being part of a method.  </doc>
Cexpr
  precompile Cblock parent
      method Cblock body
{
  [parent add [self precompile]];
  = [void_expr precompile parent method body];
}

end;


/******************** ForeignExpression ********************/

implementation class
ForeignExpression: Expression

end;

implementation instance
ForeignExpression
{
  <doc> The textual representation of the foreign (C) expression.  </doc>
  String expression;
}

<doc> Designated initializer.  </doc>
id
  initWithExpression String e
		type Type t
pre
  t != nil
{
  expression = e;
  = [super initWithType t];
}

Cexpr
  precompile Cblock parent
      method Cblock body
{
  = [[Cconst alloc self] init expression];
}

end;

/******************** Named ********************/

implementation class
Named extension tesla: tesla

end;

implementation instance
Named extension tesla

<doc> Emit the {name} to the stream {s}, replacing any nasty character
    that C does not allow in an identifier by an underscore (_).  </doc>
OutputStream (s)
  safeCName OutputStream s
{
  //
  = [s print name];
}

end;

/******************** NumberConstant ********************/

implementation class
NumberConstant extension tesla: tesla

end;

implementation instance
NumberConstant extension tesla

<doc> Return {TRUE}.  </doc>
boolean
  isConstant
{
  = TRUE;
}

Cexpr
  precompile Cblock parent
      method Cblock body
{
  = [[Cconst alloc self] init [[MutableByteString new] print value]];
}

OutputStream (s)
  writeCFile OutputStream s
{
  [s print value];
}

end;

/******************** Proxy ********************/

implementation class
Proxy: Common, tesla
{
  <doc> All currently employed proxies.  </doc>
  static MutableEqSet proxies;
}

<doc> Add the new proxy to the {proxies}.  </doc>
instance (id) (proxy)
 alloc
{
  proxy = [super alloc];
  [self addProxy proxy];
}

<doc> Add the new proxy to the {proxies}.  </doc>
instance (id) (proxy)
  alloc Common object
{
  proxy = [super alloc object];
  [self addProxy proxy];
}

<doc> Add the {proxy} to the {proxies}.  </doc>
protected void
  addProxy Proxy proxy
post
  [proxies length] == 1 + old ([proxies length])
{
  [proxies add proxy];
}

<doc> Check that there are no proxies left.  Issue an error if that is not
    the case.  </doc>
void
  checkProxies
{
  if (proxies != nil && [proxies length] > 0)
    {
      [[[ParseTom error] print ("internal error: not all proxies resolved (",
				[proxies length], " left)")] nl];
      [[[ParseTom report] print proxies] nl];
    }
}

void
  load Array arguments
{
  proxies = [MutableEqSet new];
}

end;

implementation instance
Proxy

end;

/******************** Reference ********************/

implementation class
Reference extension tesla: tesla

end;

implementation instance
Reference extension tesla

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  [referenced coarseUsage f parent c body m];
}

Cexpr
    precompile Cblock parent
	method Cblock body
{
  = [referenced precompile parent method body];
}

Cexpr (lhs)
    precompile Cblock parent
	method Cblock body
  assignmentOf Cexpr rhs
{
  lhs = [self precompile parent method body];
  [parent add [[Casgn alloc self] init (lhs, rhs)]];
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    type = [referenced force_type];
  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

OutputStream (s)
  writeCName TeslaOutput s
{
  [Variable (referenced) writeCName s];
}

end;

/******************** StringConstant ********************/

implementation class
StringConstant extension tesla: tesla

end;

implementation instance
StringConstant extension tesla

Cexpr
  precompile Cblock parent
      method Cblock body
{
  int i, n = [string length];
  MutableByteString buf = [MutableByteString withCapacity n + 2];
  [buf print "byte_string_with_c_string (\""];
  for (i = 0; i < n; i++)
    {
      byte b = string[i];
      if (b == '\\' || b == '"')
	[buf add '\\'];
      [buf add b];
    }
  [buf print "\")"];

  = [[Cconst alloc self] init buf];
}

end;

/******************** SuperReference ********************/

implementation class
SuperReference: State, tesla
{
  <doc> Mapping from a {Meta} to an array of {SuperReference}s.  The
      key is the subclass involved.  </doc>
  static MutableEqDictionary references;

  <doc> The references that are actually used.  </doc>
  public static MutableEqSet used_references;
}

<doc> Designated allocator.  </doc>
instance (id) (sref)
  from Meta subclass
    to Meta superclass
pre
  [subclass isDirectSubclassOf superclass]
{
  MutableArray a;
  if (!references)
    {
      references = [MutableEqDictionary new];
      used_references = [MutableEqSet new];
    }
  else
    a = references[subclass];

  if (!a)
    {
      a = [MutableObjectArray withCapacity 2];
      references[subclass] = a;
    }
  else
    {
      // Slow?
      // Mon Jul 13 14:27:16 1998, tiggr@gerbil.org
      int i;
      for (i = [a length] - 1; i >= 0; i--)
	{
	  sref = a[i];
	  if (superclass == [sref referenced])
	    return;
	}
    }

  sref = [[self alloc] initWithSuper superclass sub subclass];
  [a add sref];
}

end;

implementation instance
SuperReference
{
  <doc> The subclass doing the referencing.  </doc>
  Meta cer;

  <doc> The superclass being referenced.  This is direct super of
      {cer}.  </doc>
  public Meta referenced;
}

<doc> Designated initializer.  </doc>
protected id
  initWithSuper Meta ced
	    sub Meta referencer
{
  (cer, referenced) = (referencer, ced);
  = [super init];
}

<doc> Emit a declaration of this super reference to the resolver
    header file.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
  // void?
  // Fri Jul 10 11:51:43 1998, tiggr@gerbil.org
  [[self writeCReferenceName [s print "extern void * "]] print ";\n"];
}

<doc> Emit a definition of this super reference to the resolver C
    file.  </doc>
OutputStream (s)
  writeCDefinition OutputStream s
{
  [[self writeCReferenceName [s print "void * "]] print " = &"];
  [[referenced writeCMetaDefinitionName s] print ";\n"];
}

<doc> Emit this super reference and add {self} to the
    {used_references}.  </doc>
TeslaOutput (s)
  writeCFile TeslaOutput s
{
  [used_references add self];
  [self writeCReferenceName s];
}

<doc> Emit a struct initializer for this super reference.  </doc>
OutputStream (s)
  writeCInitializer OutputStream s
{
  [self writeCReferenceName [s print "{(void *) &"]];
  [cer writeCMetaReferenceName [s print ", (void *) &"]];
  [[referenced writeCMetaReferenceName [s print ", (void *) &"]] print "},\n"];
}

<doc> Emit something like {_sr_foo_bar} or {_sr_unit_foo_bar}.  </doc>
OutputStream (s)
  writeCReferenceName OutputStream s
{
  [s print "_sr_"];
  if (single_unit != nil)
    [[C write s CName [single_unit name]] print '_'];
  [referenced writeCReferenceTypeName [[cer writeCReferenceTypeName s]
					print "_"]];
}

end;

/******************** TypeDescription ********************/

implementation class
TypeDescription: State, tesla
{
  <doc> Various flags.  </doc>
  const TD_USED = 1;
  const TD_INCLUDES_DYNAMIC = 2;

  <doc> All type descriptions, keyed on their full name (which is a
      concatenation of the {TD_*} type description characters).  </doc>
  static MutableDictionary descriptions;

  <doc> All used type descriptions.  </doc>
  public static MutableEqSet used_descriptions;

  <doc> Return the TRT enum name, keyed on type description.  </doc>
  static MutableMapped trt_names;
}

<doc> Create the {descriptions}, to avoid checks in the other class
    methods.  </doc>
void
  load Array arguments
{
  descriptions = [MutableDictionary new];
  used_descriptions = [MutableEqSet new];

  /* Set up the void thing.  */
  TypeDescription td = [[self alloc] initWithName ""];
  descriptions[[td name]] = td;
  descriptions[[UniqueString with "v"]] = td;
}

void
   registerNames Array names
  ofDescriptions Array type_descriptions
{
  int i, n = [names length];
  trt_names = [MutableEqDictionary new];
  for (i = 0; i < n; i++)
    trt_names[type_descriptions[i]] = names[i];
}

<doc> Return the type description with the indicated {name}, creating it
    if necessary.  The {name}, when used to create a new instance, is
    uniqued.  </doc>
instance (id) (result)
  withName String name
{
  result = descriptions[name];
  if (!result)
    {
      result = [[self alloc] initWithName name];
      descriptions[[result name]] = result;
    }
}

end;

implementation instance
TypeDescription
{
  <doc> The so-called name of this description; it is a concatenation of
      the {TD_*} type description characters </doc>
  public String name;

  <doc> Various flags.  </doc>
  int td_flags;
}

<doc> Designated initializer.  </doc>
protected id
  initWithName String nm
{
  name = [UniqueString with nm];

  int i, n = [name length];
  for (i = 0; i < n; i++)
    if (name[i] == 'x')
      {
	td_flags |= TD_INCLUDES_DYNAMIC;
	break;
      }

  = [super init];
}

boolean
  includesDynamic
{
  = td_flags & TD_INCLUDES_DYNAMIC != 0;
}

void
  noteUsage
{
  if (td_flags & TD_USED == 0)
    {
      td_flags |= TD_USED;
      [used_descriptions add self];
    }
}

<doc> Return the TRT enum name of the receiving description.  </doc>
String (result)
  trtName
pre
  [name length] == 1
post
  result != nil
{
  = trt_names[self];
}

<doc> Declare the receiving type description.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
  [[[self writeCReferenceName [s print "extern struct trtd_selector_args "]]
     print ';'] nl];
}

<doc> Define the receiving type description.  </doc>
OutputStream (s)
  writeCDefinition OutputStream s
{
  [[self writeCReferenceName [s print "struct trtd_selector_args "]]
    print " = {"];
  int i, n = [name length];
  [s print (n, ", {")];
  for (i = 0; i < n; i++)
    {
      if (i > 0)
	[s print ", "];
      // Closed naming!
      // Fri May 29 15:11:51 1998, tiggr@gerbil.org
      byte b = name[i];
      if (b == 'r')
	[s print "TRT_TE_REFERENCE"];
      else if (b == 'v')
	[s print "TRT_TE_VOID"];
      else if (b == 'o')
	[s print "TRT_TE_BOOLEAN"];
      else if (b == 'b')
	[s print "TRT_TE_BYTE"];
      else if (b == 'c')
	[s print "TRT_TE_CHAR"];
      else if (b == 'i')
	[s print "TRT_TE_INT"];
      else if (b == 'l')
	[s print "TRT_TE_LONG"];
      else if (b == 'f')
	[s print "TRT_TE_FLOAT"];
      else if (b == 'd')
	[s print "TRT_TE_DOUBLE"];
      else if (b == 'p')
	[s print "TRT_TE_POINTER"];
      else if (b == 's')
	[s print "TRT_TE_SELECTOR"];
      else if (b == 'x')
	[s print "TRT_TE_DYNAMIC"];
    }
  [[s print "}};"] nl];
}

<doc> Emit something like {rr}.  </doc>
OutputStream (s)
  writeCName OutputStream s
{
  [C write s CName name];
}

<doc> Emit the name of the struct containing the definition of this
    selector, i.e. something like {_atd_rr} or {_atd_foo_rr}.  </doc>
OutputStream (s)
  writeCReferenceName OutputStream s
{
  [s print "_atd_"];
  if (single_unit != nil)
    [[C write s CName [single_unit name]] print '_'];
  [C write s CName name];
}

end;

/******************** ValueReferenceProxy ********************/

implementation class
ValueReferenceProxy: Proxy, Expression

end;

implementation instance
ValueReferenceProxy
{
  <doc> The name being referenced.  </doc>
  String name;

  <doc> The context of this ValueReference reference.  </doc>
  Meta context;
}

<doc> Designated initializer.  </doc>
id
  initWithName String n
     inContext Meta c
{
  (name, context) = (n, c);
  = [super initWithType nil];
}

<doc> Return another proxy.  </doc>
id
  again
{
  = [[ValueReferenceProxy alloc self] initWithName name inContext context];
}

<doc> Resolve the receiving proxy into a reference and let it deduce its
    type.  </doc>
(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
post
  [resolved type] != nil
{
  Instance context_instance;
  Expression original;

  if (type != nil)
    [self unimplemented cmd message: "type"];
  if (![proxies memq self])
    [[[ParseTom warning self] print ("not in proxies: ", self)] nl];

  Attribute a = [context attributeNamed name
			 including (TRUE, TRUE, TRUE, TRUE, TRUE, FALSE)];
  if (!a && ![context isClass])
    {
      context_instance = Instance (context);
      a = [[context itsClass] attributeNamed name
			      including (TRUE, TRUE, TRUE, TRUE, TRUE, FALSE)];
    }

  if (!a)
    {
      [[[ParseTom error self]
	 print ("reference to undeclared `", name, "'")] nl];
      type = bottom_type;
      return [super resolve (expected, convertible) context context_type
		    nested: (indices, index)];
    }

  original = [a reference context_instance];
  = [original resolve (expected, convertible) context context_type
	      nested: (indices, index)];

  /* We no longer exist.  */
  [proxies remove self];
}

<doc> Accept it (do nothing).  </doc>
void
  setIsLhs
{
}

<doc> Also emit the {name}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" name=", name)];
}

end;
