<copyright> TIG extensions class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: extensions.t,v 1.2 1999/10/12 21:24:32 tiggr Exp $</id>
    </copyright>

<c>
#include <sys/stat.h>
</c>

/******************** Unit (tig) ********************/

implementation class
mu.Unit extension tig

end;

implementation instance
mu.Unit extension tig
{
  <doc> Posing associations.  Each element is a {Cons} dotted pair
      {(sub . super)}.  </doc>
  MutableArray posers;
}

void
  loadFiles
{
  [files makeElementsPerform selector (void loadFile)];
}

<doc> Be notified that the poser {msub} poses as the posed {msuper}.
    </doc>
redefine void
  poser Meta msub
  poses Meta msuper
{
  if (!posers)
    posers = [MutableObjectArray withCapacity 1];
  [posers add [Cons with (msub, msuper)]];
}

void
  setPosers
{
  if (posers != nil)
    {
      int i, n = [posers length];
      Meta poser, posed;

      for (i = 0; i < n; i++)
	{
	  (poser, posed) = [posers[i] decons];
	  [posed setPoser poser];
	}
    }
}

end;

/******************** TomFile (tig) ********************/

implementation class
mu.TomFile extension tig: tig

end;

implementation instance
mu.TomFile extension tig
{
  public String interface_filename;
}

<doc> If necessary or requested, generate the interface file of the
    receiving TOM file.  </doc>
void
  loadFile
{
  MutableByteString ifn = [MutableByteString new];
  String bn = [File basename filename without-extension: ".t"];

  interface_filename = [ifn print (bn, ".j-tig")];

  if (flag_force_output
      || [File file-needs-updating interface_filename
	       with-respect-to filename])
    {
      LexTom l = [[LexTom alloc]
		   initWithStream [File open filename input: YES
					flags: FILE_NOT_EXIST_RAISE]
		   named filename];
      ParseTom p = [ParseTom parser self using l];
      [p parse];
      [l close];
    }
}

end;

/******************** File (tesla) ********************/

// This is a copy from the same code in Tesla, and should really belong in
// a library common to tesla and tig.
// Mon Oct 11 22:00:51 1999, tiggr@gerbil.org
implementation class
File extension tesla

<doc> Return {TRUE} if the {output} is outdated with respect to the
    {input}, i.e., if {output} does not exist or is older than {input}.
    </doc>
boolean (result)
  file-needs-updating String output
      with-respect-to String input
{
  pointer op, ip;
  int olen, ilen;

  (op, olen) = [ByteString (output) byteStringContents];
  (ip, ilen) = [ByteString (input) byteStringContents];

  <c> {
    struct stat sti, sto;
    char *os = memcpy (alloca (1 + olen), op, olen);

    os[olen] = 0;
    if (stat (os, &sto) < 0)
      result = 1;
    else
      {
	char *is = memcpy (alloca (1 + ilen), ip, ilen);
	is[ilen] = 0;

	/* No update needed if the input does not exist.  */
	if (!stat (is, &sti) && sto.st_mtime < sti.st_mtime)
	  result = 1;
      }
  } </c>
}

end;

implementation instance
File extension tesla

end;

/******************** ParseUnit (tig) ********************/

implementation class
ParseUnit extension tig

end;

implementation instance
ParseUnit extension tig

void
  notePoser Meta msub
      poses Meta msuper
{
  [unit poser msub poses msuper];
}

end;

/******************** ParseTom (tig) ********************/

implementation class
ParseTom: posing mu.ParseTom, tig

end;

implementation instance
ParseTom
{
  <doc> The output file of this parser.  </doc>
  File of;

  <doc> Iff set, we still need to emit the {end}.  </doc>
  boolean open;

  <doc> Iff not set, no super indications have been done.  </doc>
  boolean supers;
}

<doc> Open {of} before parsing and close it afterwards.  </doc>
void
  parse
{
  of = [File open [file interface_filename] output: YES
	     flags: FILE_NOT_EXIST_CREATE | FILE_EXIST_TRUNCATE];
  unwind ([of close])
    ({
      [super parse];
      [self finishExtension];
    });
}

<doc> Emit the end of the extension, if we are {open}.  </doc>
void
  finishExtension
{
  if (open)
    {
      [of print "\nend;\n"];
      open = FALSE;
    }
}

boolean
  outputSingleQualifier String q
		  first boolean first
{
  if (!first)
    [of print ' '];
  [of print q];
}

OutputStream (s)
  outputQualifiers int q
{
  boolean first = YES;

  if ([Qualified isDeferred q])
    first = [self outputSingleQualifier "deferred" first first];
  if ([Qualified isLocal q])
    first = [self outputSingleQualifier "local" first first];
  if ([Qualified isStatic q])
    first = [self outputSingleQualifier "static" first first];
  if ([Qualified isPublic q])
    first = [self outputSingleQualifier "public" first first];
  if ([Qualified isPrivate q])
    first = [self outputSingleQualifier "private" first first];
  if ([Qualified isProtected q])
    first = [self outputSingleQualifier "protected" first first];
  if ([Qualified isMutable q])
    first = [self outputSingleQualifier "mutable" first first];
  if ([Qualified isRedeclare q])
    first = [self outputSingleQualifier "redeclare" first first];
  if ([Qualified isRedefine q])
    first = [self outputSingleQualifier "redefine" first first];

  if (!first)
    [of print ' '];
}

void
  finishMethod CustomMethod m
{
  Indexed np = [m name_parts], args = [m arguments];
  int i, j, l, n, nnp = [np length], nap = !args ? 0 : [args length];

  [of print '\n'];
  [self outputQualifiers [m qualifiers]];
  [[m return_type] writeTom of];

  ArgumentTuple rn = [m return_names];
  if (rn != nil)
    [rn writeTom [of print ' ']];

  for (i = 0; i < nnp; i++)
    {
      n = [np[i] length];
      if (n > l)
	l = n;
    }

  for ({l += 2; i = 0;}; i < nnp; i++)
    {
      String sn = np[i];
      n = [np[i] length];

      [of print '\n'];
      for (j = n; j < l; j++)
	[of print ' '];
      [of print sn];
      if (i < nap)
	{
	  ArgumentSingle arg = args[i];

	  [of print ' '];
	  [[arg type] writeTom of];
	  [of print ' '];
	  [arg writeTom of];

	  Expression e = [arg default_value];
	  if (e != nil)
	    [e writeTom [of print " = "]];
	}
    }

  Expression e = [m precondition];
  if (e != nil)
    [e writeTom [of print "\npre\n  "]];

  e = [m postcondition];
  if (e != nil)
    [e writeTom [of print "\npost\n  "]];
  
  [of print ";"];
}

/********** ParseTom methods **********/

void
  startExtension mu.Extension ext
       interface boolean is_interface
{
  Meta m = [ext meta];
  boolean is_class = [m isClass];
  mu.Unit muu = [m unit];

  [self finishExtension];
  open = TRUE;

  [of print ("interface ", is_class ? "class" : "instance", "\n")];
  if (muu != top_unit)
    [of print ([muu name], ".")];

  [of print [m name]];
  if (![ext isMainExtension])
    [of print (" extension ", [ext name])];
}

void
  startSupers
{
  supers = FALSE;
}

void
  addSuper Meta meta
     shift boolean instance_only
{
  [of print supers ? ", " : ": "];
  supers = TRUE;

  mu.Unit muu = [meta unit];
  if (instance_only)
    [[meta writeTom [of print "instance ("]] print ")"];
  else
    {
      boolean class_p = [current_meta isClass];
      Meta pm = meta;

      if (class_p)
	pm = [pm meta];
      if ([current_meta posed] == pm)
	[of print "posing "];

      [meta writeTom of];
    }
}

void
  startMetaVars
{
  [of print "\n{"];
}

void
  finishMetaVars
{
  [of print "\n}"];
}

void
  addConst String name
     value Expression e
{
  [[e writeTom [of print ("\n  const ", name, " = ")]] print ';'];
}

Expression
   monop int op
    with All shs
  postp: boolean postfix = NO
pre
  shs != nil
{
  = [[UnresolvedOperatorInvocation alloc]
      initWithReceiver nil arguments [MutableObjectArray with shs] operator op
      postfixed: postfix];
}

Expression
  binop int op
   with (All, All) (lhs, rhs)
pre
  lhs != nil && rhs != nil
{
  = [[UnresolvedOperatorInvocation alloc]
      initWithReceiver nil arguments [MutableObjectArray with (lhs, rhs)]
      operator op];
}

Expression
  triop int op
   with (All, All, All) (lhs, mhs, rhs)
pre
  lhs != nil && mhs != nil && rhs != nil
{
  = [[UnresolvedOperatorInvocation alloc]
      initWithReceiver nil arguments [MutableObjectArray with (lhs, mhs, rhs)]
      operator op];
}

<doc> Return a {Named} thing acting to be an {Expression}.  </doc>
Expression
  attributeNamed String name
{
  = Any ([[Named alloc] initWithName name]);
}

void
  finishMethodDecl CustomMethod m
{
  [self finishMethod m];
}

void
  startMethodBody CustomMethod m
{
  [self finishMethod m];
  skips = TRUE;
}

void
  finishMethodBody CustomMethod m
{
  skips = FALSE;
}

void
  metaVar (int, Type, String) (qualifiers, type, name)
{
  [of print "\n  "];
  [self outputQualifiers qualifiers];
  [type writeTom of];
  [of print (' ', name, ';')];
}

end;

/******************** mu (tig) ********************/

implementation class
mu extension tig

end;

implementation instance
mu extension tig

OutputStream (s)
  writeTom OutputStream s
{
  [s print ("(* UNIMPLEMENTED writeTom ", self, " *)")];
}

end;

/******************** NumberConstant (tig) ********************/

implementation class
NumberConstant extension tig

end;

implementation instance
NumberConstant extension tig

OutputStream (s)
  writeTom OutputStream s
{
  [s print value];
}

end;

/******************** UnresolvedInvocation (tig) ********************/

implementation class
UnresolvedInvocation extension tig

end;

implementation instance
UnresolvedInvocation extension tig

OutputStream (s)
  writeTom OutputStream s
{
  int i, nn = [name_parts length], na = !arguments ? 0 : [arguments length];

  [receiver writeTom [s print '[']];
  for (i = 0; i < nn; i++)
    {
      String np = name_parts[i];

      [s print (' ', np)];
      if (i < na)
	[arguments[i] writeTom [s print ' ']];
    }

  = [s print ']'];
}

end;

/******************** All (tig) ********************/

implementation class
All extension tig

end;

implementation instance
All extension tig

// Many writeTom methods in this file come straight from tm...
// Tue Oct 12 21:32:25 1999, tiggr@gerbil.org
OutputStream
  writeTom OutputStream s
  priority int p
{
  = [self writeTom s];
}

<doc> Print {self} to {s}.  </doc>
OutputStream
  writeTom OutputStream s
{
  = [s print self];
}

end;

/******************** UnresolvedOperatorInvocation (tig) ********************/

implementation class
UnresolvedOperatorInvocation extension tig

OutputStream (s)
  write_argument (OutputStream, Argument, int) (s, arg, priority)
{
  if (!arg)
    = [arg writeTom s];
  else
    = [arg writeTom s priority priority];
}

end;

implementation instance
UnresolvedOperatorInvocation extension tig

OutputStream (s)
  writeTom OutputStream s
{
  [self writeTom s priority -1];
}

OutputStream (s)
  writeTom OutputStream s
  priority int priority
{
  int n = [arguments length];

  if (priority > operator)
    s = [s print '('];

  if (n == 1)
    {
      s = [operator_names[operator] write s];
      if (operator == OP_OLD || operator == OP_RET)
	s = [s print ' '];
      write_argument (s, arguments[0], operator);
    }
  else if (n == 2)
    {
      write_argument (s, arguments[0], operator);
      if (operator == OP_AT)
	{
	  s = [s print '['];
	  write_argument (s, arguments[1], operator);
	  s = [s print ']'];
	}
      else
	{
	  s = [s print ' '];
	  s = [operator_names[operator] write s];
	  s = [s print ' '];
	  write_argument (s, arguments[1], operator);
	}
    }
  else
    {
      write_argument (s, arguments[0], operator);
      s = [" ? " write s];
      write_argument (s, arguments[1], operator);
      s = [" : " write s];
      write_argument (s, arguments[2], operator);
    }

  if (priority > operator)
    s = [s print ')'];

  = s;
}

end;

/******************** Named ********************/

implementation class
Named extension tig

end;

implementation instance
Named extension tig

OutputStream (s)
  writeTom OutputStream s
{
  [s print name];
}

end;

/******************** Tuple ********************/

implementation class
Tuple extension tig

end;

implementation instance
Tuple extension tig

OutputStream (s)
  writeTom OutputStream s
{
  int i, n = [self length];
  [s print '('];
  for (i = 0; i < n; i++)
    {
      if (i > 0)
	[s print ", "];
      [self[i] writeTom s];
    }
  [s print ')'];
}

end;

/******************** VariableTuple ********************/

implementation class
VariableTuple extension tig

end;

implementation instance
VariableTuple extension tig

<doc> Behave as a {Tuple}.  </doc>
OutputStream (s)
  writeTom OutputStream s
{
  = [super (Tuple) writeTom s];
}

end;

/******************** TypeTuple ********************/

implementation class
TypeTuple extension tig

end;

implementation instance
TypeTuple extension tig

<doc> Behave as a {Tuple}.  </doc>
OutputStream (s)
  writeTom OutputStream s
{
  = [super (Tuple) writeTom s];
}

end;

/******************** Meta ********************/

implementation class
Meta extension tig

end;

implementation instance
Meta extension tig

OutputStream (s)
  writeTom OutputStream s
{
  mu.Extension x = [self extensionNamed nil];
  boolean class_p = [self isClass];
  TomFile f = [x file];

  if (class_p)
    [s print "class ("];

  mu.Unit u = !f ? nil : [f unit];
  if (f != nil && u != [mu.Unit builtin_unit] && u != [tig top_unit])
    [s print ([u name], '.', [self name])];
  else
    [s print [self name]];

  if (class_p)
    [s print ')'];
}

end;

/******************** StringConstant (tig) ********************/

implementation class
StringConstant extension tig

end;

implementation instance
StringConstant extension tig

OutputStream (s)
  writeTom OutputStream s
{
  [s print self];
}

end;
